# -*- coding: utf-8 -*-
#
# Copyright 2018-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Logging in to a Renku deployment."""

import os
import sys
import urllib
import uuid
import webbrowser

import git
import requests

from renku.core import errors
from renku.core.incubation.command import Command
from renku.core.models.enums import ConfigFilter
from renku.core.utils import communication
from renku.core.utils.git import get_renku_repo_url
from renku.core.utils.urls import get_remote, parse_authentication_endpoint

CONFIG_SECTION = "http"
RENKU_BACKUP_PREFIX = "renku-backup"


def login_command():
    """Return a command for logging in to the platform."""
    return Command().command(_login)


def _login(client, endpoint, git_login, yes):
    parsed_endpoint = _parse_endpoint(client, endpoint)

    remote_name, remote_url = None, None
    if git_login:
        if not client.repo:
            raise errors.ParameterError("Cannot use '--git' flag outside a project.")

        remote_name, remote_url = get_remote(client.repo)
        if remote_name and remote_url:
            if not yes:
                communication.confirm("Remote URL will be changed. Do you want to continue?", abort=True, warning=True)
        else:
            raise errors.ParameterError("Cannot find a unique remote URL for project.")

    cli_nonce = str(uuid.uuid4())

    communication.echo(f"Please log in at {parsed_endpoint.geturl()} on your browser.")

    login_url = _get_url(parsed_endpoint, "/api/auth/login", cli_nonce=cli_nonce)
    webbrowser.open_new_tab(login_url)

    server_nonce = communication.prompt("Once completed, enter the security code that you receive at the end")
    cli_token_url = _get_url(parsed_endpoint, "/api/auth/cli-token", cli_nonce=cli_nonce, server_nonce=server_nonce)

    try:
        response = requests.get(cli_token_url)
    except requests.RequestException as e:
        raise errors.OperationError("Cannot get access token from remote host.") from e

    if response.status_code == 200:
        access_token = response.json().get("access_token")
        _store_token(client, parsed_endpoint.netloc, access_token)

        if git_login:
            _store_git_credential_helper(client, parsed_endpoint.netloc)
            _swap_git_remote(client, parsed_endpoint, remote_name, remote_url)
    else:
        communication.error(
            f"Remote host did not return an access token: {parsed_endpoint.geturl()}, "
            f"status code: {response.status_code}"
        )
        sys.exit(1)


def _parse_endpoint(client, endpoint):
    parsed_endpoint = parse_authentication_endpoint(client=client, endpoint=endpoint)
    if not parsed_endpoint:
        raise errors.ParameterError("Parameter 'endpoint' is missing.")

    return parsed_endpoint


def _get_url(parsed_endpoint, path, **query_args):
    query = urllib.parse.urlencode(query_args)
    return parsed_endpoint._replace(path=path, query=query).geturl()


def _store_token(client, netloc, access_token):
    client.set_value(section=CONFIG_SECTION, key=netloc, value=access_token, global_only=True)
    os.chmod(client.global_config_path, 0o600)


def _store_git_credential_helper(client, netloc):
    client.repo.git.config("credential.helper", f"!renku token --hostname {netloc}", local=True)


def _swap_git_remote(client, parsed_endpoint, remote_name, remote_url):
    backup_remote_name = f"{RENKU_BACKUP_PREFIX}-{remote_name}"

    if backup_remote_name in [r.name for r in client.repo.remotes]:
        communication.echo(f"Backup remove '{backup_remote_name}' already exists. Ignoring '--git' flag.")
        return

    new_remote_url = get_renku_repo_url(remote_url, deployment_hostname=parsed_endpoint.netloc)

    try:
        client.repo.create_remote(backup_remote_name, url=remote_url)
    except git.GitCommandError:
        communication.error(f"Cannot create backup remote '{backup_remote_name}' for '{remote_url}'")
    else:
        try:
            client.repo.git.remote("set-url", remote_name, new_remote_url)
        except git.GitCommandError:
            client.repo.delete_remote(backup_remote_name)
            raise errors.GitError(f"Cannot change remote url for '{remote_name}' to {new_remote_url}")


def read_renku_token(client, endpoint):
    """Read renku token from renku config file."""
    try:
        parsed_endpoint = _parse_endpoint(client, endpoint)
    except errors.ParameterError:
        return
    if not parsed_endpoint:
        return

    return _read_renku_token_for_hostname(client, parsed_endpoint.netloc)


def _read_renku_token_for_hostname(client, hostname):
    return client.get_value(section=CONFIG_SECTION, key=hostname, config_filter=ConfigFilter.GLOBAL_ONLY)


def logout_command():
    """Return a command for logging out from the platform."""
    return Command().command(_logout)


def _logout(client, endpoint):
    if endpoint:
        parsed_endpoint = parse_authentication_endpoint(client=client, endpoint=endpoint)
        key = parsed_endpoint.netloc
    else:
        key = "*"

    client.remove_value(section=CONFIG_SECTION, key=key, global_only=True)
    _remove_git_credential_helper(client)
    _restore_git_remote(client)


def _remove_git_credential_helper(client):
    try:
        client.repo.git.config("credential.helper", local=True, unset=True)
    except git.exc.GitCommandError:  # NOTE: If already logged out, ``git config --unset`` raises an exception
        pass


def _restore_git_remote(client):
    if not client.repo:  # Outside a renku project
        return

    backup_remotes = [r.name for r in client.repo.remotes if r.name.startswith(RENKU_BACKUP_PREFIX)]
    for backup_remote in backup_remotes:
        remote_name = backup_remote.replace(f"{RENKU_BACKUP_PREFIX}-", "")
        remote_url = client.repo.remotes[backup_remote].url

        try:
            client.repo.git.remote("set-url", remote_name, remote_url)
        except git.GitCommandError:
            raise errors.GitError(f"Cannot restore remote url for '{remote_name}' to {remote_url}")

        try:
            client.repo.delete_remote(backup_remote)
        except git.GitCommandError:
            communication.error(f"Cannot delete backup remote '{backup_remote}'")


def token_command():
    """Return a command as git credential helper."""
    return Command().command(_token)


def _token(client, command, hostname):
    if command != "get":
        return

    # NOTE: hostname comes from the credential helper we set up and has proper format
    hostname = hostname or ""
    token = _read_renku_token_for_hostname(client, hostname) or ""

    communication.echo("username=renku")
    communication.echo(f"password={token}")



# -*- coding: utf-8 -*-
#
# Copyright 2018-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represent run templates."""

import copy
import itertools
from datetime import datetime
from pathlib import PurePosixPath
from typing import Any, Dict, List
from uuid import uuid4

from marshmallow import EXCLUDE
from werkzeug.utils import secure_filename

from renku.core.models.calamus import JsonLDSchema, Nested, fields, prov, renku, schema
from renku.core.models.entities import Entity
from renku.core.models.workflow import parameters as old_parameter
from renku.core.models.workflow.parameter import (
    CommandInput,
    CommandInputSchema,
    CommandOutput,
    CommandOutputSchema,
    CommandParameter,
    CommandParameterSchema,
)
from renku.core.models.workflow.run import Run
from renku.core.utils.urls import get_host

MAX_GENERATED_NAME_LENGTH = 25


class Plan:
    """Represent a `renku run` execution template."""

    def __init__(
        self,
        *,
        parameters: List[CommandParameter] = None,
        command: str = None,
        description: str = None,
        id: str,
        inputs: List[CommandInput] = None,
        invalidated_at: datetime = None,
        keywords: List[str] = None,
        name: str = None,
        outputs: List[CommandOutput] = None,
        success_codes: List[int] = None,
    ):
        self.command: str = command
        self.description: str = description
        self.id: str = id
        self.inputs: List[CommandInput] = inputs or []
        self.invalidated_at: datetime = invalidated_at
        self.keywords: List[str] = keywords or []
        self.name: str = name
        self.outputs: List[CommandOutput] = outputs or []
        self.parameters: List[CommandParameter] = parameters or []
        self.success_codes: List[int] = success_codes or []

        if not self.name:
            self.name = self._get_default_name()

    def __repr__(self):
        return self.name

    @classmethod
    def from_run(cls, run: Run, hostname: str):
        """Create a Plan from a Run."""
        assert not run.subprocesses, f"Cannot create a Plan from a Run with subprocesses: {run._id}"

        def extract_run_uuid(run_id: str) -> str:
            # https://localhost/runs/723fd784-9347-4081-84de-a6dbb067545b/
            return run_id.rstrip("/").rsplit("/", maxsplit=1)[-1]

        uuid = extract_run_uuid(run._id)
        plan_id = cls.generate_id(hostname=hostname, uuid=uuid)

        def convert_argument(argument: old_parameter.CommandArgument) -> CommandParameter:
            """Convert an old CommandArgument to a new CommandParameter."""
            assert isinstance(argument, old_parameter.CommandArgument)

            return CommandParameter(
                default_value=argument.value,
                description=argument.description,
                id=CommandParameter.generate_id(plan_id=plan_id, postfix=PurePosixPath(argument._id).name),
                label=None,
                name=argument.name,
                position=argument.position,
                prefix=argument.prefix,
            )

        def convert_input(input: old_parameter.CommandInput) -> CommandInput:
            """Convert an old CommandInput to a new CommandInput."""
            assert isinstance(input, old_parameter.CommandInput)

            return CommandInput(
                default_value=input.consumes.path,
                description=input.description,
                id=CommandInput.generate_id(plan_id=plan_id, postfix=PurePosixPath(input._id).name),
                label=None,
                mapped_to=input.mapped_to,
                name=input.name,
                position=input.position,
                prefix=input.prefix,
            )

        def convert_output(output: old_parameter.CommandOutput) -> CommandOutput:
            """Convert an old CommandOutput to a new CommandOutput."""
            assert isinstance(output, old_parameter.CommandOutput)

            return CommandOutput(
                create_folder=output.create_folder,
                default_value=output.produces.path,
                description=output.description,
                id=CommandOutput.generate_id(plan_id=plan_id, postfix=PurePosixPath(output._id).name),
                label=None,
                mapped_to=output.mapped_to,
                name=output.name,
                position=output.position,
                prefix=output.prefix,
            )

        return cls(
            command=run.command,
            description=run.description,
            id=plan_id,
            inputs=[convert_input(i) for i in run.inputs],
            keywords=run.keywords,
            name=run.name,
            outputs=[convert_output(o) for o in run.outputs],
            parameters=[convert_argument(a) for a in run.arguments],
            success_codes=run.successcodes,
        )

    @staticmethod
    def generate_id(hostname: str, uuid: str) -> str:
        """Generate an identifier for Plan."""
        uuid = uuid or str(uuid4())
        return f"https://{hostname}/plans/{uuid}"

    def _get_default_name(self) -> str:
        name = "-".join(str(a) for a in self.to_argv())
        if not name:
            return uuid4().hex[:MAX_GENERATED_NAME_LENGTH]

        name = secure_filename(name)
        rand_length = 5
        return f"{name[:MAX_GENERATED_NAME_LENGTH - rand_length -1]}-{uuid4().hex[:rand_length]}"

    def assign_new_id(self):
        """Assign a new UUID.

        This is required only when there is another plan which is exactly the same except the parameters' list.
        """
        current_uuid = self._extract_uuid()
        new_uuid = str(uuid4())
        self.id = self.id.replace(current_uuid, new_uuid)
        self.parameters = copy.deepcopy(self.parameters)
        for a in self.parameters:
            a.id = a.id.replace(current_uuid, new_uuid)

    def _extract_uuid(self) -> str:
        path_start = self.id.find("/plans/")
        return self.id[path_start + len("/plans/") :]

    def is_similar_to(self, other: "Plan") -> bool:
        """Return true if plan has the same inputs/outputs/arguments as another plan."""

        def get_input_patterns(plan: Plan):
            return {e.default_value for e in plan.inputs}

        def get_output_patterns(plan: Plan):
            return {e.default_value for e in plan.outputs}

        def get_parameters(plan: Plan):
            return {(a.position, a.prefix, a.default_value) for a in plan.parameters}

        # TODO: Check order of inputs/outputs/parameters as well after sorting by position
        return (
            self.command == other.command
            and set(self.success_codes) == set(other.success_codes)
            and get_input_patterns(self) == get_input_patterns(other)
            and get_output_patterns(self) == get_output_patterns(other)
            and get_parameters(self) == get_parameters(other)
        )

    def to_argv(self) -> List[Any]:
        """Convert a Plan into argv list."""
        arguments = itertools.chain(self.inputs, self.outputs, self.parameters)
        arguments = filter(lambda a: a.position is not None, arguments)
        arguments = sorted(arguments, key=lambda a: a.position)

        argv = self.command.split(" ") if self.command else []
        argv.extend(e for a in arguments for e in a.to_argv())

        return argv

    def to_run(self, client, entities_cache: Dict[str, Entity]) -> Run:
        """Create a Run."""
        uuid = self._extract_uuid()
        host = get_host(client)
        # TODO: This won't work if plan_id was randomly generated; for PoC it's OK.
        run_id = f"https://{host}/runs/{uuid}"

        def get_entity(path: str) -> Entity:
            entity = entities_cache.get(path)
            if not entity:
                entity = Entity.from_revision(client=client, path=path, revision="HEAD")
                entities_cache[path] = entity
            return entity

        def convert_parameter(argument: CommandParameter) -> old_parameter.CommandArgument:
            return old_parameter.CommandArgument(
                description=argument.description,
                id=argument.id.replace(self.id, run_id),
                label=None,
                name=argument.name,
                position=argument.position,
                prefix=argument.prefix,
                value=argument.default_value,
            )

        def convert_input(input: CommandInput) -> old_parameter.CommandInput:
            return old_parameter.CommandInput(
                consumes=get_entity(input.default_value),
                description=input.description,
                id=input.id.replace(self.id, run_id),
                label=None,
                mapped_to=input.mapped_to,
                name=input.name,
                position=input.position,
                prefix=input.prefix,
            )

        def convert_output(output: CommandOutput) -> old_parameter.CommandOutput:
            return old_parameter.CommandOutput(
                create_folder=output.create_folder,
                description=output.description,
                id=output.id.replace(self.id, run_id),
                label=None,
                mapped_to=output.mapped_to,
                name=output.name,
                position=output.position,
                prefix=output.prefix,
                produces=get_entity(output.default_value),
            )

        return Run(
            arguments=[convert_parameter(p) for p in self.parameters],
            command=self.command,
            id=run_id,
            inputs=[convert_input(i) for i in self.inputs],
            outputs=[convert_output(o) for o in self.outputs],
            successcodes=self.success_codes,
        )


class PlanSchema(JsonLDSchema):
    """Plan schema."""

    class Meta:
        """Meta class."""

        rdf_type = [prov.Plan, schema.Action, schema.CreativeWork]
        model = Plan
        unknown = EXCLUDE

    command = fields.String(renku.command, missing=None)
    description = fields.String(schema.description, missing=None)
    id = fields.Id()
    inputs = Nested(renku.hasInputs, CommandInputSchema, many=True, missing=None)
    invalidated_at = fields.DateTime(prov.invalidatedAtTime, add_value_types=True)
    keywords = fields.List(schema.keywords, fields.String(), missing=None)
    name = fields.String(schema.name, missing=None)
    outputs = Nested(renku.hasOutputs, CommandOutputSchema, many=True, missing=None)
    parameters = Nested(renku.hasArguments, CommandParameterSchema, many=True, missing=None)
    success_codes = fields.List(renku.successCodes, fields.Integer(), missing=[0])



# -*- coding: utf-8 -*-
#
# Copyright 2018-2021- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Clone a Renku repo along with all Renku-specific initializations."""

import os
from pathlib import Path

from git import GitCommandError, Repo

from renku.core import errors
from renku.core.management.githooks import install
from renku.core.models.git import GitURL


def _handle_git_exception(e, raise_git_except, progress):
    """Handle git exceptions."""
    if not raise_git_except:
        lines = progress.other_lines + progress.error_lines if progress else []
        error = "".join([f"\n\t{line}" for line in lines if line.strip()])
        message = f"Cannot clone remote Renku project: Git exited with code {e.status} and error message:\n {error}"
        raise errors.GitError(message)

    raise e


def clone(
    url,
    path=None,
    install_githooks=True,
    install_lfs=True,
    skip_smudge=True,
    recursive=True,
    depth=None,
    progress=None,
    config=None,
    raise_git_except=False,
    checkout_rev=None,
):
    """Clone Renku project repo, install Git hooks and LFS."""
    from renku.core.management.client import LocalClient

    path = path or GitURL.parse(url).name

    if isinstance(path, Path):
        path = str(path)

    # Clone the project
    if skip_smudge:
        os.environ["GIT_LFS_SKIP_SMUDGE"] = "1"

    clone_config = None
    if isinstance(config, str):
        clone_config = config
        config = None

    try:
        # NOTE: Try to clone, assuming checkout_rev is a branch (if it is set)
        repo = Repo.clone_from(
            url, path, branch=checkout_rev, recursive=recursive, depth=depth, progress=progress, config=clone_config
        )
    except GitCommandError as e:
        # NOTE: clone without branch set, in case checkout_rev was not a branch but a tag or commit
        if not checkout_rev:
            _handle_git_exception(e, raise_git_except, progress)

        try:
            repo = Repo.clone_from(url, path, recursive=recursive, depth=depth, progress=progress, config=clone_config)
        except GitCommandError as e:
            _handle_git_exception(e, raise_git_except, progress)

        try:
            # NOTE: Now that we cloned successfully, try to checkout the checkout_rev
            repo.git.checkout(checkout_rev)
        except GitCommandError as e:
            msg = str(e)
            if "is not a commit and a branch" in msg and "cannot be created from it" in msg:
                return repo, False  # NOTE: Project has no commits to check out

            raise

    try:
        repo.head.commit
    except ValueError:
        # NOTE: git repo has no head commit, which means it is empty/not a renku project
        return repo, False

    if config:
        config_writer = repo.config_writer()

        for key, value in config.items():
            key_path = key.split(".")
            key = key_path.pop()

            if not key_path or not key:
                raise errors.GitError("Cannot write to config. Section path or key is invalid.")

            config_writer.set_value(".".join(key_path), key, value)

        config_writer.release()

    client = LocalClient(path)

    if install_githooks:
        install(client=client, force=True)

    if install_lfs:
        command = ["git", "lfs", "install", "--local", "--force"]
        if skip_smudge:
            command += ["--skip-smudge"]
        try:
            repo.git.execute(command=command, with_exceptions=True)
        except GitCommandError as e:
            raise errors.GitError("Cannot install Git LFS") from e

    return repo, bool(client.project)



# -*- coding: utf-8 -*-
#
# Copyright 2018-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represent an execution of a Plan."""

import pathlib
from datetime import datetime
from pathlib import Path
from typing import List, Optional, Union
from urllib.parse import quote, urlparse
from uuid import uuid4

from git import Git, GitCommandError
from marshmallow import EXCLUDE

from renku.core.models.calamus import JsonLDSchema, Nested, fields, oa, prov, renku
from renku.core.models.cwl.annotation import Annotation, AnnotationSchema
from renku.core.models.entities import Collection, CollectionSchema, Entity, EntitySchema
from renku.core.models.provenance import qualified as old_qualified
from renku.core.models.provenance.activities import ProcessRun, WorkflowRun
from renku.core.models.provenance.agents import Person, PersonSchema, SoftwareAgent, SoftwareAgentSchema
from renku.core.models.provenance.parameter import (
    ParameterValueSchema,
    PathParameterValue,
    PathParameterValueSchema,
    VariableParameterValue,
    VariableParameterValueSchema,
)
from renku.core.models.workflow.dependency_graph import DependencyGraph
from renku.core.models.workflow.plan import Plan, PlanSchema
from renku.core.utils.urls import get_host


class Association:
    """Assign responsibility to an agent for an activity."""

    def __init__(self, *, agent: Union[Person, SoftwareAgent] = None, id: str, plan: Plan):
        self.agent: Union[Person, SoftwareAgent] = agent
        self.id: str = id
        self.plan: Plan = plan


class Usage:
    """Represent a dependent path."""

    def __init__(self, *, entity: Union[Collection, Entity], id: str):
        self.entity: Union[Collection, Entity] = entity
        self.id: str = id


class Generation:
    """Represent an act of generating a path."""

    def __init__(self, *, entity: Union[Collection, Entity], id: str):
        self.entity: Union[Collection, Entity] = entity
        self.id: str = id


class Activity:
    """Represent an activity in the repository."""

    def __init__(
        self,
        *,
        agents: List[Union[Person, SoftwareAgent]] = None,
        annotations: List[Annotation] = None,
        association: Association = None,
        ended_at_time: datetime = None,
        generations: List[Generation] = None,
        id: str,
        invalidations: List[Entity] = None,
        order: Optional[int] = None,  # TODO: Remove order and use ended_at_time for ordering
        parameters: List[Union[PathParameterValue, VariableParameterValue]] = None,
        # project=None,  # TODO: project._id gets messed up when generating and then running commands
        started_at_time: datetime = None,
        usages: List[Usage] = None,
    ):
        self.agents: List[Union[Person, SoftwareAgent]] = agents
        self.annotations: List[Annotation] = annotations or []
        self.association: Association = association
        self.ended_at_time: datetime = ended_at_time
        self.generations: List[Generation] = generations or []
        self.id: str = id
        self.invalidations: List[Entity] = invalidations or []
        self.order: Optional[int] = order
        self.parameters: List[Union[PathParameterValue, VariableParameterValue]] = parameters or []
        # self.project: Project = project
        self.started_at_time: datetime = started_at_time
        self.usages: List[Usage] = usages or []

        # TODO: _was_informed_by = attr.ib(kw_only=True)
        # TODO: influenced = attr.ib(kw_only=True)

    @classmethod
    def from_process_run(cls, process_run: ProcessRun, plan: Plan, client, order: Optional[int] = None):
        """Create an Activity from a ProcessRun."""
        hostname = get_host(client)
        activity_id = Activity.generate_id(hostname)

        association = Association(agent=process_run.association.agent, id=f"{activity_id}/association", plan=plan)

        # NOTE: The same entity can have the same id during different times in its lifetime (e.g. different commit_sha,
        # but the same content). When it gets flattened, some fields will have multiple values which will cause an error
        # during deserialization. Make sure that no such Entity attributes exists (store those information in the
        # Generation object).

        invalidations = [_convert_invalidated_entity(e, activity_id, client) for e in process_run.invalidated]
        generations = [_convert_generation(g, activity_id, client) for g in process_run.generated]
        usages = [_convert_usage(u, activity_id, client) for u in process_run.qualified_usage]

        parameters = _create_parameters(activity_id=activity_id, plan=plan, usages=usages, generations=generations)

        return cls(
            agents=process_run.agents,
            annotations=process_run.annotations,
            association=association,
            ended_at_time=process_run.ended_at_time,
            generations=generations,
            id=activity_id,
            invalidations=invalidations,
            order=order,
            parameters=parameters,
            # project=process_run._project,
            started_at_time=process_run.started_at_time,
            usages=usages,
        )

    @staticmethod
    def generate_id(hostname: str) -> str:
        """Generate an identifier for an activity."""
        # TODO: make id generation idempotent
        return f"https://{hostname}/activities/{uuid4()}"


def _convert_usage(usage: old_qualified.Usage, activity_id: str, client) -> Usage:
    """Convert an old qualified Usage to a new one."""
    commit_sha = _extract_commit_sha(entity_id=usage.entity._id)
    entity = _convert_used_entity(usage.entity, commit_sha, activity_id, client)
    assert entity, f"Top entity was not found for Usage: {usage._id}, {usage.entity.path}"

    id = f"{activity_id}/usages/{uuid4()}"

    return Usage(id=id, entity=entity)


def _convert_generation(generation: old_qualified.Generation, activity_id: str, client) -> Generation:
    """Convert an old Generation to a new one."""
    commit_sha = _extract_commit_sha(entity_id=generation.entity._id)
    entity = _convert_generated_entity(generation.entity, commit_sha, activity_id, client)
    assert entity, f"Root entity was not found for Generation: {generation._id}"

    id = f"{activity_id}/generations/{uuid4()}"

    return Generation(id=id, entity=entity)


def _convert_used_entity(entity: Entity, revision: str, activity_id: str, client) -> Optional[Entity]:
    """Convert an Entity to one with proper metadata.

    For Collections, add members that are modified in the same commit or before the revision.
    """
    assert isinstance(entity, Entity)

    checksum = _get_object_hash(revision=revision, path=entity.path, client=client)
    if not checksum:
        return None

    id_ = _generate_entity_id(entity_checksum=checksum, path=entity.path, activity_id=activity_id)

    if isinstance(entity, Collection):
        new_entity = Collection(id=id_, checksum=checksum, path=entity.path)
        for child in entity.members:
            new_child = _convert_used_entity(child, revision, activity_id, client)
            if not new_child:
                continue
            new_entity.members.append(new_child)
    else:
        new_entity = Entity(id=id_, checksum=checksum, path=entity.path)

    assert type(new_entity) is type(entity)

    return new_entity


def _convert_generated_entity(entity: Entity, revision: str, activity_id: str, client) -> Optional[Entity]:
    """Convert an Entity to one with proper metadata.

    For Collections, add members that are modified in the same commit as revision.
    """
    assert isinstance(entity, Entity)

    try:
        entity_commit = client.find_previous_commit(paths=entity.path, revision=revision)
    except KeyError:
        return None
    if entity_commit.hexsha != revision:
        return None

    checksum = _get_object_hash(revision=revision, path=entity.path, client=client)
    if not checksum:
        return None

    id_ = _generate_entity_id(entity_checksum=checksum, path=entity.path, activity_id=activity_id)

    if isinstance(entity, Collection):
        new_entity = Collection(id=id_, checksum=checksum, path=entity.path)
        for child in entity.members:
            new_child = _convert_generated_entity(child, revision, activity_id, client)
            if not new_child:
                continue
            new_entity.members.append(new_child)
    else:
        new_entity = Entity(id=id_, checksum=checksum, path=entity.path)

    assert type(new_entity) is type(entity)

    return new_entity


def _convert_invalidated_entity(entity: Entity, activity_id: str, client) -> Optional[Entity]:
    """Convert an Entity to one with proper metadata."""
    assert isinstance(entity, Entity)
    assert not isinstance(entity, Collection), f"Collection passed as invalidated: {entity._id}"

    commit_sha = _extract_commit_sha(entity_id=entity._id)
    commit = client.find_previous_commit(revision=commit_sha, paths=entity.path)
    commit_sha = commit.hexsha
    checksum = _get_object_hash(revision=commit_sha, path=entity.path, client=client)
    if not checksum:
        # Entity was deleted at commit_sha; get the one before it to have object_id
        checksum = _get_object_hash(revision=f"{commit_sha}~", path=entity.path, client=client)
        if not checksum:
            print(f"Cannot find invalidated entity hash for {entity._id} at {commit_sha}:{entity.path}")
            return

    id_ = _generate_entity_id(entity_checksum=checksum, path=entity.path, activity_id=activity_id)
    new_entity = Entity(id=id_, checksum=checksum, path=entity.path)
    assert type(new_entity) is type(entity)

    return new_entity


def _generate_entity_id(entity_checksum: str, path: Union[Path, str], activity_id: str) -> str:
    quoted_path = quote(str(path))
    path = pathlib.posixpath.join("blob", entity_checksum, quoted_path)

    return urlparse(activity_id)._replace(path=path).geturl()


def _get_object_hash(revision: str, path: Union[Path, str], client):
    try:
        return client.repo.git.rev_parse(f"{revision}:{str(path)}")
    except GitCommandError:
        # NOTE: The file can be in a submodule or it was not there when the command ran but was there when workflows
        # were migrated (this can happen only for Usage); the project might be broken too.
        return _get_object_hash_from_submodules(path, client)


def _get_object_hash_from_submodules(path: Union[Path, str], client) -> str:
    for submodule in client.repo.submodules:
        try:
            path_in_submodule = Path(path).relative_to(submodule.path)
        except ValueError:
            continue
        else:
            try:
                return Git(submodule.abspath).rev_parse(f"HEAD:{str(path_in_submodule)}")
            except GitCommandError:
                pass


def _extract_commit_sha(entity_id: str) -> str:
    # NOTE: extracts commit sha from ids like /blob/a3bf8a165dd56da078b96f2ca2ff22f14a3bdd57/input
    path = urlparse(entity_id).path
    assert path.startswith("/blob/"), f"Invalid entity identifier: {entity_id}"

    commit_sha = path[len("/blob/") :].split("/", 1)[0]
    assert len(commit_sha) == 40, f"Entity does not have valid commit SHA: {entity_id}"

    return commit_sha


def _create_parameters(activity_id, plan: Plan, usages: List[Usage], generations: List[Generation]):
    parameters = []

    inputs = {i.default_value: i for i in plan.inputs}
    for usage in usages:
        input = inputs.pop(usage.entity.path, None)
        assert input is not None, f"Cannot find usage path '{usage.entity.path}' in plan {plan.id}"
        id = PathParameterValue.generate_id(activity_id)
        parameters.append(PathParameterValue(id=id, parameter=input, path=usage.entity.path))

    assert not inputs, f"Not all inputs are converted: {inputs}"

    outputs = {o.default_value: o for o in plan.outputs}
    for generation in generations:
        output = outputs.pop(generation.entity.path, None)
        assert output is not None, f"Cannot find generation path '{generation.entity.path}' in plan {plan.id}"
        id = PathParameterValue.generate_id(activity_id)
        parameters.append(PathParameterValue(id=id, parameter=output, path=generation.entity.path))

    assert not outputs, f"Not all outputs are converted: {outputs}"

    for parameter in plan.parameters:
        id = VariableParameterValue.generate_id(activity_id)
        parameters.append(VariableParameterValue(id=id, parameter=parameter, value=parameter.default_value))

    return parameters


class ActivityCollection:
    """Equivalent of a workflow file."""

    def __init__(self, activities: List[Activity] = None):
        self.activities: List[Activity] = activities or []

    @classmethod
    def from_activity(cls, activity: Union[ProcessRun, WorkflowRun], dependency_graph: DependencyGraph, client):
        """Convert a ProcessRun/WorkflowRun to ActivityCollection."""

        def get_process_runs(workflow_run: WorkflowRun) -> List[ProcessRun]:
            # NOTE: Use Plan subprocesses to get activities because it is guaranteed to have correct order
            sorted_ids = [s.process._id for s in workflow_run.association.plan.subprocesses]
            activities = []
            # NOTE: it's possible to have subprocesses with similar ids but it does not matter since they have the same
            # plan
            # TODO: Remove these redundant subprocesses
            for id_ in sorted_ids:
                for s in workflow_run.subprocesses.values():
                    if s.association.plan._id == id_:
                        activities.append(s)
                        break
            assert len(activities) == len(workflow_run.subprocesses)
            return activities

        process_runs = get_process_runs(activity) if isinstance(activity, WorkflowRun) else [activity]

        self = ActivityCollection()

        for process_run in process_runs:
            assert isinstance(process_run, ProcessRun)
            run = process_run.association.plan
            if run.subprocesses:
                assert len(run.subprocesses) == 1, f"Run in ProcessRun has multiple steps: {run._id}"
                run = run.subprocesses[0]

            plan = Plan.from_run(run=run, hostname=get_host(client))
            plan = dependency_graph.add(plan)

            activity = Activity.from_process_run(process_run=process_run, plan=plan, client=client)
            self.add(activity)

        return self

    def add(self, activity: Activity) -> None:
        """Add an Activity."""
        self.activities.append(activity)


class AssociationSchema(JsonLDSchema):
    """Association schema."""

    class Meta:
        """Meta class."""

        rdf_type = prov.Association
        model = Association
        unknown = EXCLUDE

    agent = Nested(prov.agent, [SoftwareAgentSchema, PersonSchema])
    id = fields.Id()
    plan = Nested(prov.hadPlan, PlanSchema)


class UsageSchema(JsonLDSchema):
    """Usage schema."""

    class Meta:
        """Meta class."""

        rdf_type = prov.Usage
        model = Usage
        unknown = EXCLUDE

    id = fields.Id()
    # TODO: DatasetSchema, DatasetFileSchema
    entity = Nested(prov.entity, [EntitySchema, CollectionSchema])


class GenerationSchema(JsonLDSchema):
    """Generation schema."""

    class Meta:
        """Meta class."""

        rdf_type = prov.Generation
        model = Generation
        unknown = EXCLUDE

    id = fields.Id()
    # TODO: DatasetSchema, DatasetFileSchema
    entity = Nested(prov.qualifiedGeneration, [EntitySchema, CollectionSchema], reverse=True)


class ActivitySchema(JsonLDSchema):
    """Activity schema."""

    class Meta:
        """Meta class."""

        rdf_type = prov.Activity
        model = Activity
        unknown = EXCLUDE

    agents = Nested(prov.wasAssociatedWith, [PersonSchema, SoftwareAgentSchema], many=True)
    annotations = Nested(oa.hasTarget, AnnotationSchema, reverse=True, many=True)
    association = Nested(prov.qualifiedAssociation, AssociationSchema)
    ended_at_time = fields.DateTime(prov.endedAtTime, add_value_types=True)
    generations = Nested(prov.activity, GenerationSchema, reverse=True, many=True, missing=None)
    id = fields.Id()
    invalidations = Nested(prov.wasInvalidatedBy, EntitySchema, reverse=True, many=True, missing=None)
    order = fields.Integer(renku.order)
    parameters = Nested(
        renku.parameter,
        [ParameterValueSchema, PathParameterValueSchema, VariableParameterValueSchema],
        many=True,
        missing=None,
    )
    path = fields.String(prov.atLocation)
    started_at_time = fields.DateTime(prov.startedAtTime, add_value_types=True)
    usages = Nested(prov.qualifiedUsage, UsageSchema, many=True)



# -*- coding: utf-8 -*-
#
# Copyright 2018-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes to represent inputs/outputs/parameters in a Plan."""

import urllib
from pathlib import PurePosixPath
from typing import Any, List, Optional
from uuid import uuid4

from marshmallow import EXCLUDE

from renku.core.models.calamus import JsonLDSchema, Nested, fields, rdfs, renku, schema
from renku.core.models.workflow.parameters import MappedIOStream, MappedIOStreamSchema
from renku.core.utils.urls import get_slug

RANDOM_ID_LENGTH = 4


class CommandParameterBase:
    """Represents a parameter for a Plan."""

    def __init__(
        self,
        *,
        default_value: Any,
        description: str,
        id: str,
        label: str,
        name: str,
        position: Optional[int],
        prefix: str,
    ):
        self.default_value: Any = default_value
        self.description: str = description
        self.id: str = id
        self.label: str = label
        self.name: str = name
        self.position: Optional[int] = position
        self.prefix: str = prefix

        if not self.name:
            self.name = self._get_default_name()
        if not self.label:
            self.label = self._get_default_label()

    @staticmethod
    def _generate_id(plan_id: str, parameter_type: str, position: Optional[int], postfix: str = None) -> str:
        """Generate an id for parameters."""
        # https://localhost/plans/723fd784-9347-4081-84de-a6dbb067545b/inputs/1
        # https://localhost/plans/723fd784-9347-4081-84de-a6dbb067545b/inputs/stdin
        # https://localhost/plans/723fd784-9347-4081-84de-a6dbb067545b/inputs/dda5fcbf-0098-4917-be46-dc12f5f7b675
        position = str(position) if position is not None else str(uuid4())
        postfix = urllib.parse.quote(postfix) if postfix else position
        return f"{plan_id}/{parameter_type}/{postfix}"

    @property
    def role(self) -> str:
        """Return a unique role for this parameter within its Plan."""
        assert self.id, "Id is not set"
        return PurePosixPath(self.id).name

    def to_argv(self) -> List[Any]:
        """String representation (sames as cmd argument)."""
        if self.prefix:
            if self.prefix.endswith(" "):
                return [self.prefix[:-1], self.default_value]
            return [f"{self.prefix}{self.default_value}"]

        return [self.default_value]

    def _generate_name(self, base) -> str:
        name = get_slug(self.prefix.strip(" -=")) if self.prefix else base
        position = self.position or uuid4().hex[:RANDOM_ID_LENGTH]
        return f"{name}-{position}"

    def _get_default_label(self) -> str:
        raise NotImplementedError

    def _get_default_name(self) -> str:
        raise NotImplementedError


class CommandParameter(CommandParameterBase):
    """An argument to a command that is neither input nor output."""

    def __init__(
        self,
        *,
        default_value: Any = None,
        description: str = None,
        id: str,
        label: str = None,
        name: str = None,
        position: Optional[int] = None,
        prefix: str = None,
    ):
        super().__init__(
            default_value=default_value,
            description=description,
            id=id,
            label=label,
            name=name,
            position=position,
            prefix=prefix,
        )

    @staticmethod
    def generate_id(plan_id: str, position: Optional[int] = None, postfix: str = None) -> str:
        """Generate an id for CommandParameter."""
        return CommandParameterBase._generate_id(
            plan_id, parameter_type="parameters", position=position, postfix=postfix
        )

    def _get_default_label(self) -> str:
        return f"Command Parameter '{self.default_value}'"

    def _get_default_name(self) -> str:
        return self._generate_name(base="parameter")


class CommandInput(CommandParameterBase):
    """An input to a command."""

    def __init__(
        self,
        *,
        default_value: Any = None,
        description: str = None,
        id: str,
        label: str = None,
        mapped_to: MappedIOStream = None,
        name: str = None,
        position: Optional[int] = None,
        prefix: str = None,
    ):
        super().__init__(
            default_value=default_value,
            description=description,
            id=id,
            label=label,
            name=name,
            position=position,
            prefix=prefix,
        )
        self.mapped_to: MappedIOStream = mapped_to

    @staticmethod
    def generate_id(plan_id: str, position: Optional[int] = None, postfix: str = None) -> str:
        """Generate an id for CommandInput."""
        return CommandParameterBase._generate_id(plan_id, parameter_type="inputs", position=position, postfix=postfix)

    def to_stream_representation(self) -> str:
        """Input stream representation."""
        return f" < {self.default_value}" if self.mapped_to else ""

    def _get_default_label(self) -> str:
        return f"Command Input '{self.default_value}'"

    def _get_default_name(self) -> str:
        return self._generate_name(base="input")


class CommandOutput(CommandParameterBase):
    """An output from a command."""

    def __init__(
        self,
        *,
        create_folder: bool = False,
        default_value: Any = None,
        description: str = None,
        id: str,
        label: str = None,
        mapped_to: MappedIOStream = None,
        name: str = None,
        position: Optional[int] = None,
        prefix: str = None,
    ):
        super().__init__(
            default_value=default_value,
            description=description,
            id=id,
            label=label,
            name=name,
            position=position,
            prefix=prefix,
        )
        self.create_folder: bool = create_folder
        self.mapped_to: MappedIOStream = mapped_to

    @staticmethod
    def generate_id(plan_id: str, position: Optional[int] = None, postfix: str = None) -> str:
        """Generate an id for CommandOutput."""
        return CommandParameterBase._generate_id(plan_id, parameter_type="outputs", position=position, postfix=postfix)

    def to_stream_representation(self) -> str:
        """Input stream representation."""
        if not self.mapped_to:
            return ""

        return f" > {self.default_value}" if self.mapped_to.stream_type == "stdout" else f" 2> {self.default_value}"

    def _get_default_label(self) -> str:
        return f"Command Output '{self.default_value}'"

    def _get_default_name(self) -> str:
        return self._generate_name(base="output")


class CommandParameterBaseSchema(JsonLDSchema):
    """CommandParameterBase schema."""

    class Meta:
        """Meta class."""

        rdf_type = [renku.CommandParameterBase, schema.Property]
        model = CommandParameterBase
        unknown = EXCLUDE

    default_value = fields.Raw(schema.defaultValue, missing=None)
    description = fields.String(schema.description, missing=None)
    id = fields.Id()
    label = fields.String(rdfs.label, missing=None)
    name = fields.String(schema.name, missing=None)
    position = fields.Integer(renku.position, missing=None)
    prefix = fields.String(renku.prefix, missing=None)


class CommandParameterSchema(CommandParameterBaseSchema):
    """CommandParameter schema."""

    class Meta:
        """Meta class."""

        rdf_type = [renku.CommandParameter]
        model = CommandParameter
        unknown = EXCLUDE


class CommandInputSchema(CommandParameterBaseSchema):
    """CommandInput schema."""

    class Meta:
        """Meta class."""

        rdf_type = [renku.CommandInput]
        model = CommandInput
        unknown = EXCLUDE

    mapped_to = Nested(renku.mappedTo, MappedIOStreamSchema, missing=None)


class CommandOutputSchema(CommandParameterBaseSchema):
    """CommandOutput schema."""

    class Meta:
        """Meta class."""

        rdf_type = [renku.CommandOutput]
        model = CommandOutput
        unknown = EXCLUDE

    create_folder = fields.Boolean(renku.createFolder, missing=False)
    mapped_to = Nested(renku.mappedTo, MappedIOStreamSchema, missing=None)
